"""Config.py -- store application configuration information.

This module defines the Config class, which stores application
configuration variables. You define the list of configuration
variables when you create an instance of Config (see below).
You use the configuration variables by indexing the instance
of Config with the textual name of the configuration variable.

To create: 'Config(decl)', where 'decl' is a dictionary,
which has the names of the configuration variables as its
indexes, and tuples (type, default value, description)
describing the variable as the values. For example, to
define the variables 'keep-backups' and 'backup-suffix',
you would create a Config object like this:

    c = Config({ "keep-backups": (CONFIG_BOOLEAN, 1, "Keep backups?"),
		 "backup-suffix": (CONFIG_STRING, ".bak", "Backup suffix")})

To use the variables, you would write code like this:

    if not c["keep-backups"]:
        delete_backups()

    c["backup-suffix"] = "~"
   
The allowed types for the configuration variables are
'CONFIG_BOOLEAN' (logical values), and 'CONFIG_STRING'
(string values). Is is easy to add additional types.

'read(file, ignorebad=1)'

	Read a configuration file, and modify configuration
	variables in memory accordingly. If 'ignorebad' is
	false (zero), the 'error' exception is raised, if the
	configuration file contains an unknown variable, or
	the syntax is wrong.
	
'write(file, defaults=1)'

	Write all configuration files to a file. If
	'defaults' is false (zero), only variables with
	values different from the default value are written
	to the file.

'rewrite(oldfile, newfile, defaults=1)'

	Like 'write', but reads an existing file, and replaces
	variable assignments in it with assignments to set
	the current value. Other variables are added at the end.
	'defaults' works as for 'write'.

The configuration file format is simple: Empty lines are
ignored. Comment lines (also ignored, of course) are on a
separate line (not on the same line as assignments) and begin
with a hash sign ('#'), after optional whitespace. Other lines
are assignment lines, and consist of a variable name, an equals
('='), and the value. The value may not be divided onto several
lines (backslash-continuation does not work).

Boolean values are "0", "no", "off", "false" for false,
"1", "yes", "on", "true" for true. (Case is not significant.)

String values use a C-like notation (all standard escapes
allowed). If the value both begins and ends with a double quote
('"'), those are ignored. Leading and trailing whitespace
outside double quotes are removed.

See also ConfigEditor.py.

Lars Wirzenius, liw@iki.fi
"""

from string import *

error = 'Config.error'

CONFIG_STRING = 'CONFIG_STRING'
CONFIG_BOOLEAN = 'CONFIG_BOOLEAN'

_unescape = {
	'\\': '\\', 'a': '\a', 'b': '\b', 'f': '\f', 'n': '\n', 
	'r': '\r', 't': '\t', 'v': '\v',
}

_escape = {
	'\\': '\\\\', '\a': '\\a', '\b': '\\b', '\f': '\\f', '\n': '\\n', 
	'\r': '\\r', '\t': '\\t', '\v': '\\v',
}

def str2strval(str):
	if str[:1] == '"' and str[-1:] == '"':
		str = str[1:-1]
	prevBS = prevOCT = prevHEX = code = codelen = 0
	result = ""
	for c in str:
		if (prevOCT or prevBS) and c in octdigits:
			code = code * 8 + atoi(c, 8)
			if not prevOCT:
				result = result + chr(code)
			else:
				result = result[:-1] + chr(code)
			codelen = codelen + 1
			if codelen == 3:
				prevOCT = codelen = code = 0
			else:
				prevOCT = 1
			prevBS = 0
		elif prevOCT:
			prevOCT = codelen = code = 0
		elif prevHEX and c in hexdigits:
			code = code * 16 + atoi(c, 16)
			result = result[:-1] + chr(code)
		elif prevBS and c == 'x':
			prevHEX = 1
			prevBS = code = 0
			result = result + "x"
		elif prevBS:
			try: result = result + _unescape[c]
			except KeyError: pass
			prevBS = prevHEX = prevOCT = code = codelen = 0
		elif c == '\\':
			prevBS = 1
			prevHEX = prevOCT = code = codelen = 0
		else:
			result = result + c
			prevBS = prevHEX = prevOCT = 0
	return result


def str2boolval(str):
	str = lower(str)
	if str in ["yes", "true", "on", "1"]:
		return 1
	if str in ["no", "false", "off", "0"]:
		return 0
	return None

_str2val = {
	CONFIG_STRING: str2strval,
	CONFIG_BOOLEAN: str2boolval,
}


def strval2str(str):
	result = ""
	for c in str:
		if c in _escape.keys():
			c = _escape[c]
		result = result + c
	return result

def boolval2str(bool):
	if bool:
		return "yes"
	else:
		return "no"

_val2str = {
	CONFIG_STRING: strval2str,
	CONFIG_BOOLEAN: boolval2str,
}

def isEmpty(line):
	line = strip(line)
	return line == "" or line[0] == "#"

class Config:
	error = "Config.error"

	def __init__(self, decl):
		self.values = {}
		self.decl = decl
	
	def __len__(self):
		return len(self.decl)
	
	def __getitem__(self, key):
		try:
			return self.values[key]
		except KeyError:
			pass
		return self.decl[key][1]
	
	def __setitem__(self, key, value):
		self.values[key] = value
	
	def __delitem__(self, key):
		del self.values[key]
		del self.decl[key]
	
	def keys(self):
		return self.decl.keys()

	def varStr(self, name):
		value = self[name]
		return _val2str[self.varType(name)](value)
	
	def varType(self, key):
		return self.decl[key][0]
	
	def varDesc(self, key):
		return self.decl[key][2]
	
	def varDefault(self, key):
		return self.decl[key][1]
		
	def parse(self, line, lineno, ignorebad):
		line = lstrip(line)
		x = split(line, "=", 1)
		if len(x) != 2:
			raise error, "Syntax error, line %d: " \
				"missing equals (=)" % (lineno)
		try:
			name = strip(x[0])
			decl = self.decl[name]
			value = _str2val[decl[0]](strip(x[1]))
			if value == None:
				raise error, \
				  "Invalid value on line %d" % (lineno)
		except KeyError:
			if ignorebad:
				return None, None
			raise error, "Unknown variable `%s' " \
				"on line %d" % (name, lineno)
		return name, value

	def read(self, file, ignorebad=1):
		lineno = 0
		for line in file.readlines():
			lineno = lineno + 1
			if isEmpty(line):
				continue
			name, value = self.parse(line, lineno, ignorebad)
			if name != None:
				self.values[name] = value

	def writeVar(self, new, name, defaults):
		if defaults or name in self.values.keys():
			new.write("%s = %s\n" % (name, self.varStr(name)))

	def write(self, new, defaults=1):
		for name in self.decl.keys():
			self.writeVar(new, name, defaults)
			
	def rewrite(self, old, new, ignorebad=1, defaults=1):
		lineno = 0
		written = {}
		for line in old.readlines():
			lineno = lineno + 1
			if isEmpty(line):
				new.write(line + "\n")
				continue
			name, value = self.parse(line, lineno, ignorebad)
			if name != None:
				self.writeVar(new, name, defaults)
				written[name] = 1
		for name in self.decl.keys():
			if not name in written.keys():
				self.writeVar(new, name, defaults)
			
if __name__ == "__main__":
	c = Config({ "foo": (CONFIG_STRING, "foo", "The foo variable"),
		     "bar": (CONFIG_BOOLEAN, 1, "The bar variable") })
	print "foo", c["foo"]
	c["foo"] = "foo2"
	print "foo", c["foo"]
	print "bar", c["bar"]

	oldfile = 'foo = "fo\\xA\\ bo3"\nnotexist=1'
	from StringIO import StringIO
	c.read(StringIO(oldfile))
	print "foo", c["foo"]
	print "bar", c["bar"]

	oldfile = 'foo = "fo\\xA\\ bo3"\nnotexist=1'
	newfile = StringIO()
	c.rewrite(StringIO(oldfile), newfile, defaults=0)
	newfile.seek(0,0)
	print newfile.readlines()
