"""ConfigEditor.py -- GUI configuration/option editor widget

This module defines the ConfigEditor class, which lets users
edit variables stored in instances of Config (see Config.py).

'ConfigEditor(config, layout)'

	Create a configuration editor. 'config' is an instance
	of 'Config'. 'layout' which defines the layout for
	the editor.
	
	The main editor window consists of a list of "panes",
	and "OK" and "Cancel" buttons. Clicking on the name of
	one of the panes gives you a new window, called pane,
	which some subset of the configuration variables.
	'layout' defines what panes exist, and what variables
	are on each pane.
	
	For example, if you want two panes, with "spam" and
	"food" in one, and "hostname" and "port" in the other,
	the call would be:
	
		ConfigEditor(cfg, [["Edibles", "spam", "food"],
				   ["Networking", "hostname", "port"]])
	
	'layout' is a list of lists. The element lists contain
	strings; the first string defines the name of the pane,
	the rest define the variables that belong on the pane.
	The pane will show the variables in the order given.

	When the user presses "Cancel", the editor just vanishes.
	When the user presses "OK", the instance of Config is
	modified accordingly.

This module requires the Shells module, from UITools by Mitch Chapman.

Lars Wirzenius, liw@iki.fi
"""

from Tkinter import *
from Config import *
import Shells

class LabeledWidget(Frame):
	def __init__(self, master, label, width):
		Frame.__init__(self, master)
		self.pack(expand=YES, fill=X)
		x = Label(self, text=label, width=width, anchor=W)
		x.pack(side=LEFT)

class TextField(LabeledWidget):
	def __init__(self, master, label, width, value):
		LabeledWidget.__init__(self, master, label, width)
		self.widget = Entry(self, width=40)
		self.widget.pack(side=LEFT, expand=YES, fill=X)
		self.widget.insert('end', value)
	
	def get(self):
		return self.widget.get()
	
class BooleanField(LabeledWidget):
	def __init__(self, master, label, width, value):
		LabeledWidget.__init__(self, master, label, width)
		self.var = BooleanVar()
		self.widget = Checkbutton(self, command=self.setOnOffText,
						variable=self.var)
		self.widget.pack(side=LEFT)
		self.var.set(value)
		self.setOnOffText()
	
	def setOnOffText(self):
		if self.var.get() == 0:
			text = "No"
		else:
			text = "Yes"
		self.widget.config(text=text)
	
	def get(self):
		return self.var.get()
	
class Pane(Shells.NonModal):
	def __init__(self, master, title, items, config):
		Shells.NonModal.__init__(self, master)
		self.top.title(title)
		self.top.iconname(title)
		
		longest = ""
		for item in items:
			d = config.varDesc(item)
			if len(longest) < len(d):
				longest = d
		w = len(longest)

		self.fields = {}
		for item in items:
			t = config.varType(item)
			d = config.varDesc(item)
			v = config[item]
			if t == CONFIG_STRING:
				f = TextField(self.top, d, w, v)
			elif t == CONFIG_BOOLEAN:
				f = BooleanField(self.top, d, w, v)
			else:
				raise TypeError, "unknown config var type"
			self.fields[item] = f
		
		self.config = config
		b = Button(self.top, text="Dismiss", command=self.terminate)
		b.pack(side=LEFT)
		
	def save(self):
		for field in self.fields.keys():
			self.config[field] = self.fields[field].get()
	
class ConfigEditor(Frame):
	def __init__(self, config, layout):
		self.top = Toplevel()
		self.top.title("Configuration Editor")
		self.top.iconname("Configuration Editor")
		
		Frame.__init__(self, master=self.top)
		self.pack(expand=YES, fill=BOTH)

		self.config = config
		
		self.panes = []
		for x in layout:
			p = Pane(self, x[0], x[1:], config)
			self.panes.append(p)
			b = Button(self, text=x[0], command=p.show, anchor=W)
			b.pack(side=TOP, fill=X)
		
		Button(self.top, text="OK", command=self.ok).pack(side=LEFT)
		Button(self.top, text="Cancel", command=self.cancel).pack(side=LEFT)
	
	def ok(self):
		for pane in self.panes:
			pane.save()
		self.quit()
	
	def cancel(self):
		self.quit()

	def quit(self):
		self.top.destroy()

if __name__ == "__main__":
	c = Config({
		"moderator": (CONFIG_STRING, 'liw@iki.fi', 'Moderator address'),
		"nntp-server": (CONFIG_STRING, 'localhost', 'News server'),
		"use-pgp-moose": (CONFIG_BOOLEAN, 0, 'Use PGP Moose'),
		})
	def doit():
		ce = ConfigEditor(c, [
			["Addresses", "moderator"],
			["News", "nntp-server", "use-pgp-moose"]
			])
	Button(None, text="ConfigEditor", command=doit).pack()
	mainloop()
	for i in c.keys():
		print i, c[i]
